//Timecode types
var ZEROS = "00:00:00.00";
var INFINITY = "24:00:00.00";
var HIDDEN_TIME = "99:99:99.99";
var MIN_BLINK_TIME = 0.01; //default interval between outtime and next intime, in seconds

//Alignments
var ALIGN_LEFT = 1;
var ALIGN_CENTER = 2;
var ALIGN_RIGHT = 3;
var ALIGN_TOP = 1;
var ALIGN_BOTTOM = 2;

var CONFIG_FILE = "cadet.config";

/// global configuration
var version = "1.0.01";
var fs = require('fs');
var url = require('url');
var mime = require('mime');
var himalaya = require('himalaya');//for reading ttml as html
var extend = require('extend');//for cloning an object (in JS, direct assignment is by reference)
var open = require('open');



///////  START FILEBROWSER DECLARES //////
var http = require('http');
var _ = require('lodash');
var express = require('express');
var fs = require('fs');
var path = require('path');
var util = require('util');
var os = require('os');

var program = require('commander');

function collect(val, memo) {
  if(val && val.indexOf('.') != 0) val = "." + val;
  memo.push(val);
  return memo;
}


program
  .option('-p, --port <port>', 'Port to run the file-browser. Default value is 8000')
  .option('-e, --exclude <exclude>', 'File extensions to exclude. To exclude multiple extension pass -e multiple times. e.g. ( -e .js -e .cs -e .swp) ', collect, [])
  .option('-d, --dir <dir>', 'Home directory', collect, [])
  .parse(process.argv);

var webroot=""; //= program.dir.toString().substr(1);
//console.log("SERVER: parsing commandline: webroot=["+webroot+"]");

if( !loadConfig() ){
	if( program.dir.toString().trim().length > 0 ){
		webroot = program.dir.toString().substr(1);
		//console.log("!loadConfig, program.dir exists: webroot=["+webroot+"]");
	} else {
		webroot = "/";
		program.dir = webroot;
		//console.log("!loadConfig, NO program.dir exists: webroot=["+webroot+"]");
	}
}

/*
var app = express();
var dir =  webroot;
program.dir = webroot;
*/
dir = webroot;
program.dir = webroot;


/**/
//console.log("file-browser section: begin dir analysis: program.dir=["+program.dir+"]");
if(!program.dir){
	//console.log("FILEBROWSER: !program.dir: dir=["+dir+"]");
	//console.log("file-browser: NO other user dir specified, using ["+dir+"]");
} else {
	//console.log("FILEBROWSER: program.dir=["+program.dir+"]");
	var foo = (program.dir.toString().substr(0,1)=="." ? program.dir.toString().substr(1) : program.dir.toString());  //fix up program.dir, seems to have a "." prepended
	dir=foo;
	program.dir = dir;
	//console.log("FILEBROWSER: after fixup: program.dir=["+program.dir+"]");
	//console.log("FILEBROWSER: after fixup: dir=["+dir+"]");
} 
//console.log("FILEBROWSER: dir=["+dir+"]");



///////  END FILEBROWSER DECLARES //////


/// command line arguments
var optimist = require("optimist");
var argv = optimist
    .options({
        port: {
            description: "TCP Port",
            alias: "p",
            default: 8000,
        },
        reserve: {
            describe: "reserve a channel- double quotes surround channel and commas separate components",
            alias: "r"
        },
        logLevel: {
            describe: "loggin level - error, info, connection, debug",
            alias: ["l", "log"],
            default: "info"
        },
        sslPort: {
            default: 0
        },
        help: {
            describe: "show this help message",
            alias: "h"
        }
    }).argv;

if (argv["help"]) {
    optimist.showHelp();
    process.exit(1);
} // if

console.log("NodeJS process PID=["+process.pid+"]");

var log = initializeLogger(argv["logLevel"]);
log.info(versionInfo());
//log.info(getDateTime());

var port = argv["port"];
//console.log("using port=["+port+"]");
//var sslPort = argv["sslPort"];
//var reservedChannels = argv["reserve"] || "";


var serverStatisticsUpdate = null; // used to cancel setTimeout

var _ = require("underscore");
var express = require("express");
var app = express();

/*
var bodyParser = require('body-parser');
app.use(bodyParser.json()); // support json encoded bodies
app.use(bodyParser.urlencoded({ extended: true })); // support encoded bodies
*/
var bodyParser = require('body-parser');
app.use(bodyParser.json({limit: '50mb'}));
app.use(bodyParser.urlencoded({limit: '50mb', extended: true}));


var http = require("http");
var server = http.createServer(app);

Number.prototype.toTimecode = function (n, comma) {
	if( this == -1 ) return " ";
	//var sec_num = parseInt(this, 10); // don't forget the second param (radix 10)
	var sec_num = this;
	var hours   = Math.floor(sec_num / 3600);
	var minutes = Math.floor((sec_num - (hours * 3600)) / 60);
	//var seconds = (sec_num - (hours * 3600) - (minutes * 60)).format(2);
	var seconds = (sec_num - (hours * 3600) - (minutes * 60)).toFixed(n);

	if (hours   < 10) {hours   = "0"+hours;}
	if (minutes < 10) {minutes = "0"+minutes;}
	if (seconds < 10) {seconds = "0"+seconds;}
	var time    = hours+':'+minutes+':'+seconds;
	if( comma )
		time = time.replace(".",",");
	return time;
}
String.prototype.timecodeToNumber = function () {
				if( this.trim().length==0 )	//blank timecode represented as -1 numerically
					return -1;
			
	//var sec_num = parseInt(this, 10); // don't forget the second param (radix 10)
	//01234567890
	//hh:mm:ss.ss
	var sec_num = parseFloat(this.substr(6,this.length));
	var hours   = parseFloat(this.substr(0,2))*3600;
	var minutes = parseFloat(this.substr(3.5))*60;
	//var seconds = (sec_num - (hours * 3600) - (minutes * 60)).format(2);
	return hours+minutes+sec_num;
}


//delete all .temp.* files from /public/clips directory

server.listen(port, function() {
    log.info("Server listening on port " + server.address().port);
    log.info("File browsering root: " + webroot);
}).on('error', function(){	//assume CADET is already running on the port
	log.info("Error listening on port:"+port);
	return;
});
open("http://localhost:"+port+"/");


/// routes

//////////////////////////
// ADD HEADERS
//////////////////////////
app.use(function (req, res, next) {

    // Website you wish to allow to connect
    res.setHeader('Access-Control-Allow-Origin', 'http://localhost:8000');
    res.setHeader('Access-Control-Allow-Origin', 'http://127.0.0.1:8000');
    res.setHeader('Access-Control-Allow-Origin', 'http://localhost');
    res.setHeader('Access-Control-Allow-Origin', 'http://127.0.0.1');

    // Request methods you wish to allow
    res.setHeader('Access-Control-Allow-Methods', 'GET, POST, OPTIONS, PUT, PATCH, DELETE');

    // Request headers you wish to allow
    res.setHeader('Access-Control-Allow-Headers', 'X-Requested-With,content-type');

    // Set to true if you need the website to include cookies in the requests sent
    // to the API (e.g. in case you use sessions)
    res.setHeader('Access-Control-Allow-Credentials', true);

    // Pass to next layer of middleware
    next();
});


//////////////////////////
// MAIN ENTRY FILE ROUTES
//////////////////////////
app.get('/', function(req, res){
    //log.debug("app.get /:  sending cadet.html ");
    res.sendFile('cadet.html', {"root": __dirname} );
});
app.get("/cadet", function(req, res, next) {
    //log.debug("routes:all: /cadet");
    res.sendFile("cadet.html", {"root": __dirname});
}); // app.get
app.get("/cadet.html", function(req, res, next) {
    //log.debug("routes:all: /cadet.html");
    res.sendFile("cadet.html", {"root": __dirname});
}); // app.get
app.post("/quit/", function(req, res, next) {
    log.debug("/quit/ called");
    server.close();
    process.exit(0)
}); // app.post /quit/
/*************
****************/


////////////////////////////////
// ROUTE CONFIGEDIT 
////////////////////////////////
app.get("/configedit/:rand", function(req, res) {
	log.debug("app.get(//configedit//): ENTER");
    // each key in req.body will match the keys in the data object that you passed in

	var thedata = sendConfigHtml(res);	
	log.debug("app.post(//config//): thedata=["+thedata+"]");
    
	//res.send("SERVER:configedit cadet.config=["+thedata+"]");
	return;
});//app.post("/configedit/"

//config.html is set up to allow user to edit cadet configuration
function sendConfigHtml(res){
	//log.debug("sendConfigHtml: enter");
	
	//log.debug("sendConfigHtml: __dirname=["+__dirname+"]");
	//var thefile = __dirname+"/cadet.config";
	var thefile = __dirname+"/config.html";
	//log.debug("sendConfigHtml: thefile=["+thefile+"]");
	
	fs.readFile(thefile, 'utf8', function (err,data) {
	  if (err) {
	  	//log.debug("sendConfigHtml: Error reading file=["+thefile+"]");
		return log.debug(err);
	  }
	  //log.debug("sendConfigHtml: data=["+data+"]");
	  data = data.replace("%%webroot%%",webroot);
	  res.send( data);
	});
}//sendConfigHtml

// CONFIG-SAVE
app.post("/configsave/", function(req, res) {
	//log.debug("app.post(//configsave//): ENTER");
    // each key in req.body will match the keys in the data object that you passed in

	var myObject = JSON.parse(req.body.configdefaults);
	log.debug("app.post(//configsave//): configdefaults.defaults.webroot=["+JSON.stringify(myObject.defaults.webroot)+"]");
	if( !dirExists(myObject.defaults.webroot) ){
		log.debug("app.post(//configsave//): webroot=["+myObject.defaults.webroot+"] doesn't exist, returning");
			res.status(400).send("Server says webroot dir=["+myObject.defaults.webroot+"] doesn't exist");
		return;
	}
    
	var wstream = fs.createWriteStream(__dirname + "/" + CONFIG_FILE);
	wstream.write(JSON.stringify(myObject,null,2));
	wstream.end();
	
	//save it locally
	webroot = myObject.defaults.webroot;
	dir = webroot;//??
	res.status(200).send("OK");
	return;
});//app.post("/configsave/"
//config.html is set up to allow user to edit cadet configuration

function loadConfig(){
	//try to load webroot from config file, if possible; return true if successful
	//log.debug("loadConfig: enter");

	var thefile = __dirname + "/" + CONFIG_FILE;
	//console.log("loadConfig: thefile=["+thefile+"]");

	try{
		data = fs.readFileSync(thefile, "utf8");
		var obj = JSON.parse(data);
		//console.log("loadConfig: from file json webroot=["+obj.defaults.webroot+"]");
		if( obj.defaults && obj.defaults.webroot ){
			//console.log("loadConfig: setting webroot=["+obj.defaults.webroot+"]");
			webroot = obj.defaults.webroot.toString();
		} else {
			//console.log("loadConfig: obj.defaults doesn't exist or obj.defaults.webroot doesn't exist");
		}

		//console.log("loadConfig: finally webroot=["+webroot+"]");
		return true;
	} catch (err){
		console.log("loadConfig: Error reading file=["+thefile+"]");
		console.log("loadConfig: err=["+err+"]");
		return false;
	}
}//loadConfig


// CONFIG-READ
app.post("/configread/", function(req, res) {
	//log.debug("app.post(//configread//): ENTER");
	res.send( readConfig() );
	return;
});//app.post("/configread/"
function readConfig(){
	//log.debug("readConfig: enter");

	var thefile = __dirname + "/" + CONFIG_FILE;
	//console.log("readConfig: thefile=["+thefile+"]");

	try{
		data = fs.readFileSync(thefile, "utf8");
		var obj = JSON.parse(data);
		//console.log("readConfig: from file json webroot=["+obj.defaults.webroot+"]");
		if( obj.defaults ){
			//console.log("readConfig: returning raw data");
			return data;
		} else {
			//console.log("readConfig: obj.defaults doesn't exist");
			return null;
		}

	} catch (err){
		//console.log("readConfig: Error reading file=["+thefile+"] err=["+err+"]  Returning NULL");
		return null;
	}
}//readConfig

////////////////////////////////
// ROUTE GETWEBROOT
////////////////////////////////
app.get("/getwebroot/", function(req, res) {
	//log.debug("app.post(//configsave//): ENTER");
    // each key in req.body will match the keys in the data object that you passed in
	//log.debug("app.get(/getwebroot/: webroot=["+webroot+"]");
	var obj = {};
	obj.webroot = webroot;
	res.send(JSON.stringify(obj));
});//app.post("/getwebroot/"


////////////////////////////////
// ROUTE CADET SUPPORT FILES 
////////////////////////////////
app.get("/-cad3t-/*", function(req, res, next) {
    //log.debug("app.get /-cad3t-/*: req=[" + req.params[0] +"]");
    
    var thefile = req.params[0];//.replace("-cad3t-/","");

    try{ 
    	//log.debug("app.get /-cad3t-/*: trying root=["+__dirname+"] thefile=["+thefile+"]");
    	fs.statSync(__dirname + "/" + thefile);
    	//log.debug("app.get /-cad3t-/*: IS A FILE IN ./");
    	if( thefile == "config.html" ){
    	} else {
    		res.sendFile(thefile, {"root": __dirname});
    	}
    }
    catch(ex){
				res.send("Error: " + req.params[0]+ " not found");
    }// catch ./
    //next();
}); // app.get


//////////////////////////
// ROUTE FILEBROWSER
//////////////////////////
var includeMap=null;
app.get('/files', function(req, res) {
 //log.debug("APP.GET /files: dir=["+dir+"]");
 if( req.query.includeMap ){
	 includeMap = JSON.parse(req.query.includeMap);
	 //log.debug("APP.GET /files: includeMap=["+JSON.stringify(includeMap)+"]");
	 //for(var i=0; i<includeMap.length; i++)
	 	//log.debug("includeMap["+i+"]=["+includeMap[i]+"]");
 }
 var oldCurrentDir =  dir;
 var currentDir =  dir;
 currentDir =  dir;
 //log.debug("app.get /files: oldCurrentDir=["+oldCurrentDir+"]");
 var query = req.query.path || '';
 //log.debug("app.get /files: desiredDir=["+query+"]");
 if (query) currentDir = path.join(dir, query);
 //log.debug("app.get /files: currentDir=["+currentDir+"]");
 //log.debug("app.get /files: browsing currentDir=["+currentDir+"]");
 try{
	 //fixup symbolic links like "My Documents" in Windows
	 currentDir = fs.realpathSync(currentDir)
 
	 fs.readdir(currentDir, function (err, files) {
		 if (err) {
			console.log("Error trying to scan directory = ["+currentDir+"]");
			//throw err;
			currentDir = oldCurrentDir;
			return;
		  }
		  var data = [];
		  files
		  .filter(function (file) {
			  return true;
		  }).forEach(function (file) {
			try {
					//console.log("/files/: processing file=["+file+"]");
				
					//ignore hidden/special files on mac
					if( file.startsWith(".") || file.startsWith("$") ){
						return;
					}
				  
					var isDirectory = fs.statSync(path.join(currentDir,file)).isDirectory();
					if (isDirectory) {
					  //console.log("/files: isDirectory=true file=["+file+"]");
					  data.push({ Name : file, IsDirectory: true, Path : path.join(query, file)  });
					} else {
					  //console.log("/files: isDirectory=false file=["+file+"]");
					  //check for excludes
					  var ext = path.extname(file);
					  //console.log("/files: isDirectory=false ext=["+ext+"]");
					  if(program.exclude && _.contains(program.exclude, ext)) {
						//console.log("excluding file ", file);
						return;
					  }
					  //check for includes 

					  if( includeMap[0]=="*" ){
						//console.log("/files: include due to *");
						data.push({ Name : file, Ext : ext, IsDirectory: false, Path : path.join(query, file) });
					  }else if( ext && _.contains(includeMap, ext) ){
						//console.log("/files: contains includeMap");
						data.push({ Name : file, Ext : ext, IsDirectory: false, Path : path.join(query, file) });
					  }
					}

			} catch(e) {
			  console.log(e); 
			}        
		
		  });
		  data = _.sortBy(data, function(f) { return f.Name });
		  res.json(data);
		  //console.log("APP.GET /files: jsondata=["+JSON.stringify(data)+"]");
	  });
  }//try watching for errors like inaccessible directories like "My Documents" in Windows
  catch(e) {
  	//console.log("Error on readdir");
  }
});


//////////////////////////
// ROUTE checkExists
//////////////////////////
app.post('/checkExists/', function(req, res) {
    //log.debug(".post(/checkExists/): ENTER");
    //log.debug(".post(/checkExists/): ENTER data passed=["+JSON.stringify(req.body,null,2)+"]");
	var myObj = JSON.parse(req.body.thedata);
    //log.debug(".post(/checkExists/): ENTER to check on ["+myObj.src+"]");
    
	try {
    	//log.debug(".post(/checkExists/): trying ["+webroot+myObj.src+"]");
		var stats = fs.statSync(webroot+myObj.src);
    	//log.debug(".post(/checkExists/): statStync says exists ["+myObj.src+"]");
		res.send("Server says ["+myObj.src+"] exists");
	}
	catch(err) {
    	//log.debug(".post(/checkExists/): statStync says doesn't exist ["+myObj.src+"]");
		res.status(400).send("Server says ["+myObj.src+"] doesn't exist");
	}
    
});//app.post('/checkExists/:src'
function dirExists(path){
	return(fs.existsSync(path));
}//dirExists

///////////////////////////
// ALL OTHER FILE ROUTES
///////////////////////////
app.get("/*", function(req, res, next) {
    //log.debug("app.get /*: req.params[0]=[" + req.params[0] +"]");

    try{ 
    	//log.debug("app.get /*: trying webroot/req.params[0]=[" + webroot + "/" + req.params[0]+"]");
    	fs.statSync(webroot + "/" + req.params[0]);
    	//log.debug("app.get /*: IS A FILE DIRECTLY");
    	res.status(200).sendFile(req.params[0], {"root": webroot + "/"});
    }
    catch(ex){
				//log.debug("app.get /*: ["+webroot + "/" + req.params[0]+"] IS NOT A FILE");
				res.send("Error: file=[" + webroot + "/" +req.params[0]+ "] not found");
    }// catch 
    //next();
}); // app.get

///////////////////////////
// ROUTE SAVE
///////////////////////////
app.post("/saveproject/", function(req, res) {
	//log.debug(".post(/saveproject/): ENTER");
    // each key in req.body will match the keys in the data object that you passed in
    var myObject = JSON.parse(req.body.thedata);
    //log.debug(".post(/saveproject/*): stringify myObject=["+JSON.stringify(myObject)+"]");
    
    //log.debug(".post(/saveproject/*): myObject.mediaFilename=["+myObject.mediaFilename+"]");
    //log.debug(".post(/saveproject/*): myObject.trackFilename=["+myObject.trackFilename+"]");
    //log.debug(".post(/saveproject/*): events.length=["+myObject.events.length+"]");
    //log.debug(".post(/saveproject/*): stringify element 0._text=["+JSON.stringify(myObject.events[0], null, 2)+"]");
    //log.debug(".post(/saveproject/*): element 0.text=["+myObject.events[0].text+"]");
    //log.debug(".post(/saveproject/*): element 0.htmlText=["+myObject.events[0].htmlText+"]");
    
    
    //save the project file
    //saveProject("myProject.cadet", myObject);
    saveProject(myObject, res);
    
    //res.send("Saved project");
    return;
    
});//app.post("/saveproject/"
function saveProject(oProject, res){
	//log.debug("saveProject: enter");

    var projName, projDir, projPath;
    
    if( oProject.config.webProjectFile.trim().length == 0 ){
    	//log.debug("saveProject: no json webprojectFILE setting to Untitled.cadet");
    	projName = "Untitled.cadet";
    } else {
    	//log.debug("saveProject: yes json webprojectFILE=["+oProject.config.webProjectFile+"]");
    	projName = oProject.config.webProjectFile;    		
    }
    //fix it up to end with .cadet
	if( !projName.endsWith(".cadet") )
		projName += ".cadet";    		

    	
    //if( oProject.config.webProjectDir.trim().length == 0 ){
    	//log.debug("saveProject: no json webprojectDIR setting to webroot");
    	//projDir = webroot;
    //} else {
    	//log.debug("saveProject: yes json webprojectDIR=["+oProject.config.webProjectDir+"]");
    	projDir = oProject.config.webProjectDir;
    //}
    	
    //if( oProject.config.webProjectPath.trim().length == 0 ){
    	//log.debug("saveProject: no json webprojectPATH setting to webroot+'/'+projName");
    	//projPath = webroot+"/"+ projDir + "/"+projName;
    	//projPath = webroot  + projDir + "/"+ projName; //webroot always ends with slash??
    	if( !webroot.endsWith("/") && !projDir.startsWith("/") && !projDir.startsWith("\\") )
    		projPath = webroot + "/" + projDir + "/"+ projName; //webroot always ends with slash??
    	else
    		projPath = webroot + projDir + "/"+ projName; //webroot always ends with slash??
    	log.info("saveProject: PATH=["+projPath+"]");
    	
    //} else {
    	//log.debug("saveProject: yes json webprojectPATH=["+oProject.config.webProjectPath+"]");
    	//projPath = oProject.config.webProjectPath;
    //}
    	
	//projFullPath = oProject.config.mediaDir+"/"+sname
    
    //log.debug("saveProject: oProject=["+JSON.stringify(oProject, null, 2)+"]");
    //log.debug("saveProject: projPath=["+projPath+"]");
	var strm = fs.createWriteStream(projPath);
	strm.write(JSON.stringify(oProject, null, 2));
	strm.end
	
	res.send("Saved project file to ["+projPath+"]");
}//saveProject

///////////////////////////
// ROUTE REFRESHCAPTIONS
///////////////////////////
app.post("/refreshCaptions*", function(req, res) {
	//log.debug("app.post(//refreshCaptions//): ENTER");
    // each key in req.body will match the keys in the data object that you passed in
    var myObject = JSON.parse(req.body.thedata);  
    
    if(myObject.config.mediaFilename.trim().length==0){
    	//log.debug("post(/refreshCaptions*: NO BASEMEDIA=["+myObject.config.mediaFilename+"]");
    	res.send("No base media file at["+myObject.config.mediaFilename+"]");
    	return;
    }
    
 	//log.debug("root __dirname=["+__dirname+"]");
	//log.debug("app.post(//refreshCaptions//): clip path=["+myObject.config.mediaDir+"/"+myObject.config.mediaFilename+"]");
    //log.debug("app.post(//refreshCaptions//): trak path=["+myObject.config.mediaDir+"/"+myObject.config.trackFilename+"]");
	//return;//!!!!!!!!!!!!!!!
	
	
    //write a VTT file using media path as the base path for track; it will always live next to the media
	
    var trackFullPath;
    
    //get base path (all except .extention) but start at homeDir/public
    var basemediapath, basemediadir;
    //webroot starts with a / but shouldn't end with one; mediaDir should begin with one since it comes from the filebrowser
    basemediadir = webroot+myObject.config.mediaDir + "/";
    basemediapath = webroot+myObject.config.mediaDir + "/" + myObject.config.mediaFilename;
    basemediapath = basemediapath.substring(0, basemediapath.lastIndexOf(".")); 
    
    //write the temp file
    trackFullPath = basemediapath+".temp.vtt";
    
    //make sure basemediapath exists
    if( !fs.existsSync(basemediadir) ){
    	log.debug("post(/refreshCaptions*: NOFIND=["+basemediadir+"]");
    	res.send("No media file at["+basemediadir+"]");
    	return;
    }
    
    
    //log.debug(".post(//refreshCaptions//): writing temp VTT file=["+trackFullPath+"]");
	writeVTT(trackFullPath,myObject);
    
	res.send("Caption file refreshed");
	return;
});//app.post("/refreshCaptions/"

//////////////////////////
// ROUTE EXPORTTYPE
//////////////////////////
app.post("/exporttype/:etype", function(req, res) {
	//log.debug("app.post(//exporttype//): ENTER");
    // each key in req.body will match the keys in the data object that you passed in
    var myObject = JSON.parse(req.body.thedata);    
    
 	//log.debug("root __dirname=["+__dirname+"]");
	//log.debug("app.post(//exporttype//): clip path=["+myObject.config.mediaDir+"/"+myObject.config.mediaFilename+"]");
    //log.debug("app.post(//exporttype//): trak path=["+myObject.config.mediaDir+"/"+myObject.config.trackFilename+"]");
    

    //get base path (all except .extention)
    var exportFilename, exportDir, exportPath;
    var basemediapath;
    
    if( myObject.config.exportFilename && myObject.config.exportDir ){
    	exportFilename = myObject.config.exportFilename;
    	exportDir = myObject.config.exportDir;
    } else if( myObject.config.mediaFilename && myObject.config.mediaDir ){
    	exportFilename = myObject.config.mediaFilename;
    	exportDir = myObject.config.mediaDir;
    } else {
    	exportFilename = "Untitled."
    	exportDir = "";
    }
    //convert all backslash to forwardslash
    exportDir = exportDir.replace(/\\/g, '/');
    log.debug("//exporttype//: webroot=["+webroot+"] exportDir=["+exportDir+"] exportFilename=["+exportFilename+"]");
    
    if( exportDir == "/" ) exportDir = "";
    //exportPath = webroot + "/" + exportDir + "/" + exportFilename;
    //strip leading/trailing slashes and insert them as necessary
    //exportPath = "/" + webroot.replace(/^\/+|\/+$/g, '') + "/" + (exportDir.length>0 ? exportDir.replace(/^\/+|\/+$/g, '') + "/" : "") + exportFilename;
	//WINDOWS TEMPLATE
	if( webroot.substr(1,1)==":" ) //don't prepend root slash
    	exportPath = webroot.replace(/^\/+|\/+$/g, '') + "/" + (exportDir.length>0 ? exportDir.replace(/^\/+|\/+$/g, '') + "/" : "") + exportFilename;
	else
    	exportPath = "/" + webroot.replace(/^\/+|\/+$/g, '') + "/" + (exportDir.length>0 ? exportDir.replace(/^\/+|\/+$/g, '') + "/" : "") + exportFilename;


    exportPath = exportPath.substring(0, exportPath.lastIndexOf(".")); 
    //log.debug("//export//: exportPath=["+exportPath+"]");
    
    /*
    //webroot starts with a / but shouldn't end with one; mediaDir should begin with one since it comes from the filebrowser
    if( !myObject.config.mediaFilename )
    	myObject.config.mediaFilename = "Untitled.";
    basemediapath = webroot+myObject.config.mediaDir + "/" + myObject.config.mediaFilename;
    basemediapath = basemediapath.substring(0, basemediapath.lastIndexOf(".")); 
    //log.debug("app.post(//exporttype:etype//): basemediapath=["+basemediapath+"]");
    
    
    exportPath = basemediapath;
    */
    
    //log.debug("app.post(//exporttype//): export type=["+req.params.etype+"]");
    var etype = req.params.etype.toString().toLowerCase();
    var trackFullPath;
	switch (etype.toLowerCase()) {
		case "vtt":
			trackFullPath = exportPath+".vtt"; //basemediapath+".vtt";
			//log.debug("app.post(//exporttype:etype//): writing VTT file=["+trackFullPath+"]");
			writeVTT(trackFullPath,myObject);
			res.send("Exported timed text type=["+etype+"]<br/>to file=["+trackFullPath+"]");
			return;
			break;
		case "srt":
			trackFullPath = exportPath+".srt";
			//log.debug("app.post(//exporttype:etype//): writing SRT file=["+trackFullPath+"]");
			writeSRT(trackFullPath,myObject);
			var thefile = exportFilename;
			thefile = exportFilename.substring(0, exportFilename.lastIndexOf("."));
			res.send("Exported timed text type=["+etype+"]<br/>to file=["+trackFullPath+"]<br/>******************<br/>Note: if you intend to upload to Facebook, rename the file to '"+thefile+".en_US.srt'<br/>******************<br/>");
			return;
			break;
		case "ttml":
			trackFullPath = exportPath+".ttml";
			//log.debug("app.post(//exporttype:etype//): writing TTML file=["+trackFullPath+"]");
			writeTTML(trackFullPath,myObject);
			res.send("Exported timed text type=["+etype+"]<br/>to file=["+trackFullPath+"]<br/>******************<br/>Note: if you intend to view in IE or Edge, make sure to set the fileType on the server to 'application/ttml+xml'<br/>******************<br/>");
			return;
			break;
		case "qttext":
			trackFullPath = exportPath+".qttext.txt";
			//log.debug("app.post(//exporttype:etype//): writing VTT file=["+trackFullPath+"]");
			writeQTtext(trackFullPath,myObject);
			res.send("Exported timed text type=["+etype+"]<br/>to file=["+trackFullPath+"]");
			return;
			break;
		case "sami":
			//console.log("export SAMI not implemented");
			trackFullPath = exportPath+".smi";
			//log.debug("app.post(//exporttype:etype//): writing SAMI file=["+trackFullPath+"]");
			writeSAMI(trackFullPath,myObject);
			res.send("Exported timed text type=["+etype+"]<br/>to file=["+trackFullPath+"]");
			return;
			break;
		case "plain":
			trackFullPath = exportPath+".txt";
			//log.debug("app.post(//exporttype:etype//): writing PLAIN TEXT file=["+trackFullPath+"]");
			writePlainText(trackFullPath,myObject);
			res.send("Exported timed text type=["+etype+"]<br/>to file=["+trackFullPath+"]");
			return;
			break;
		default:
			console.log("No valid export type specified.");
			res.send("Invalid export type specified ["+etype+"]");
			return
	}
    
    //should never get here
	res.send("Exported timed text type=["+etype+"]<br/>to file=["+trackFullPath+"]<br/>");
	return;
});//app.post("/exporttype/"


////////////////
//// END REST
////////////////



/// utility functions



/// versioning

function versionInfo() {
    var path = require("path");
    return path.basename(__filename) + " v" + (arguments[0] || version) + " ["+getDateTime()+"]";
} // versionString


/// Logging

function initializeLogger(logLevel) {
    var winston = require("winston");
    var logger = new(winston.Logger)({
        transports: [
                //new (winston.transports.File)({ filename: 'somefile.log' }),
                new(winston.transports.Console)({
                    level: logLevel
                })
            ] // transports
    }); // logger

    logger.setLevels({
        debug: 0, // noisy
        connection: 1, // connect/disconnect messages
        info: 2,
        error: 3 // only errors
    });
    return logger;
} // initializeLogger

///////////////////////////////////////////////////////
///////////////////////////////////////////////////////
// FILE EXPORT functions
///////////////////////
// EXPORT QTTEXT
///////////////////////
function writeQTtext(fpath, obj){
	var wstream = fs.createWriteStream(fpath);
	writeFileQTtext(wstream,obj)
	wstream.end();
}//writeQTtext
function writeFileQTtext(strm, o){
/*
https://developer.apple.com/library/mac/documentation/QuickTime/QT6_3/Chap1/QT6WhatsNew.html
{QTtext}{font:Geneva}{plain}{size:12}{textColor: 65535, 65535, 65535}{backColor: 0, 0, 0}{justify:center}{timeScale:600}{timeStamps:absolute}{language:0}{textEncoding:0}
[00:00:00.000]
Trip to Drumlin Farm
[00:00:02.000]
{child noise}
[00:00:04.000]
Hi kids!
[00:00:06.000]
*/

	//write header section
	strm.write('{QTtext}\n');
	strm.write('{font:Arial}{plain}{size:12}{textColor: 65535, 65535, 65535}{backColor: 0, 0, 0}{justify:center}{timeScale:600}{width:160}{height:48}{timeStamps:absolute}{language:0}{textEncoding:0}\n');
	strm.write('\n');
	
	var intime, outtime;
	for( var i=0; i<o.events.length; i++ ){
		intime = o.events[i].intime;
		outtime = o.events[i].outtime
		strm.write('\n');
		strm.write("["+o.events[i].intime.toTimecode(3) + ']\n');
		strm.write(eventToQTtext(o.events[i]) +'\n');
	}
	//write last outtime
	strm.write('\n');
	strm.write("["+outtime.toTimecode(3) + ']\n');
	
}//writeFileQTtext
function eventToQTtext(e){
	var sfoo="", ret="", tr={};
	for( var i=0; i<e.textRuns.length; i++ ){
		tr=e.textRuns[i];
		sfoo = tr.text;
		//log.debug("eventToQTtext: sfoo=["+sfoo+"]");
		if( sfoo.length==0 ) sfoo = " ";
		
		//prepend newline for this run
		if( tr.linebreak ) sfoo = "\n"; //sfoo = "\n"+sfoo;

		ret += sfoo;
	}//textRuns
	
	return ret;
}//eventToQTtext




///////////////////////
// EXPORT VTT
///////////////////////
function writeVTT(fpath, obj){
	var wstream = fs.createWriteStream(fpath);
	writeFileVTT(wstream,obj)
	wstream.end();
}//writeVTT
function writeFileVTT(strm, o){
	//log.debug("writeFileVTT: data=["+JSON.stringify(o, null, 2)+"]");
	//write header section
	strm.write('WEBVTT\n');
	strm.write('\n');
	
	//maybe don't do this for IE???
	/*
	strm.write('STYLE\n');
	strm.write('\n');
	strm.write('::cue(.red){ color: red; }\n');
	strm.write('::cue(.blue){ color: blue; }\n');
	strm.write('::cue(.green){ color: green; }\n');
	strm.write('::cue(.yellow){ color: yellow; }\n');
	strm.write('::cue(.cyan){ color: cyan; }\n');
	strm.write('::cue(.magenta){ color: magenta; }\n');
	strm.write('::cue(.bgblack){ background-color: black; }\n');
	strm.write('::cue(.bgwhite){ background-color: white; }\n');
	strm.write('::cue(.bgred){ background-color: red; }\n');
	strm.write('::cue(.bgblue){ background-color: blue; }\n');
	strm.write('::cue(.bggreen){ background-color: green; }\n');
	strm.write('::cue(.bgyellow){ background-color: yellow; }\n');
	strm.write('::cue(.bgcyan){ background-color: cyan; }\n');
	strm.write('::cue(.bgmagenta){ background-color: magenta; }\n');
	strm.write('\n');
	*/
	
	for( var i=0; i<o.events.length; i++ ){
		strm.write('\n');
		strm.write((i+1)+'\n');
		strm.write(o.events[i].intime.toTimecode(3) + ' --> ' + o.events[i].outtime.toTimecode(3) +'  align:'+itoeAlign(o.events[i].align)+'  line:'+itoeVAlign(o.events[i].valign)+'\n');
		strm.write(eventToVTT(o.events[i]) +'\n');
	}
	strm.write('\n');
}//writeFileVTT
function eventToVTT(e){
	var sfoo="", ret="", tr={}, bgottext = false;
	//log.debug("eventToVTT: textRuns=["+JSON.stringify(e.textRuns, null, 2)+"]");
	for( var i=0; i<e.textRuns.length; i++ ){
		//if( i==0 ) sfoo = "";
		tr=e.textRuns[i];
		sfoo = tr.text;
		
		/*//done on client side, in cleanUpBlankLines()
		if( i==0 ){	//first run, left trim
			sfoo = sfoo.replace(/^\s+/,"");
		}
		*/
		
		//log.debug("eventToVTT:1 tr.text=["+sfoo+"]");
		
		if( sfoo.trim().length > 0 )
			bgottext = true;
		//log.debug("eventToVTT: tr.id=["+tr.id+"] bgotext=["+bgottext+"]");
		
		
		if( tr.linebreak ){
			if( bgottext ){	//linebreak and already found text
				//log.debug("eventToVTT: tr.id=["+tr.id+"] linebreak=true, bgottext=true");
				ret += "\n";
				bgottext = false;
				continue;
			} else {		//linebreak and no text yet
				//log.debug("eventToVTT: tr.id=["+tr.id+"] linebreak=true, bgottext=false");
				continue;
			}
		} else {
		}
		
				
		//log.debug("eventToVTT: run sfoo=["+sfoo+"] len=["+sfoo.length+"] endval=["+sfoo.charCodeAt(1)+"]");
		//if( sfoo.length==0 ) sfoo = " ";//"\u200B";//use zero-width space instead of a space// =" ";
		if( i==e.textRuns.length-1 ) sfoo = sfoo.replace(/\s+$/, ''); //sfoo.trim(); //trim the last trailing spaces
		//log.debug("eventToVTT:2 sfoo=["+sfoo+"]");
		if( tr.bold ) sfoo = sfoo.engulfVTT("b");
		if( tr.italic ) sfoo = sfoo.engulfVTT("i");
		if( tr.underline ) sfoo = sfoo.engulfVTT("u");
		sfoo = sfoo.engulfColorVTT(tr.color,tr.bgcolor);
		//log.debug("eventToVTT:3 sfoo=["+sfoo+"]");
		
		ret += sfoo;
		//log.debug("eventToVTT: normal ret now=["+ret+"]");
		
	}//textRuns
	
	//log.debug("eventToVTT: finally return ret=["+ret+"] len=["+ret.length+"]");
	return ret;

}//eventToVTT
String.prototype.engulfVTT = function (tag) {
	//wrap the string in <tag></tag>
	var sfoo = this;
	return("<"+tag+">"+this+"</"+tag+">"); 
}
String.prototype.engulfColorVTT = function (fgc,bgc) {
	//wrap the string in <c.fgc.bgc></c>
	var sfoo = this;
	var fgname=rgbToName(fgc), bgname=rgbToName(bgc);
	return("<c."+rgbToName(fgc)+".bg"+rgbToName(bgc)+">"+this+"</c>"); 
}


///////////////////////
// EXPORT SRT
///////////////////////
function writeSRT(fpath, obj){
	var wstream = fs.createWriteStream(fpath);
	writeFileSRT(wstream,obj)
	wstream.end();
}//writeSRT
function writeFileSRT(strm, o){
	//no header
	
	for( var i=0; i<o.events.length; i++ ){
		strm.write('\n');
		strm.write((i+1)+'\n'); //don't put in subtitle numbers for SRT, youtube doesn't like them
		strm.write(o.events[i].intime.toTimecode(3,true) + ' --> ' + o.events[i].outtime.toTimecode(3,true)+'\n');
		strm.write(eventToSRT(o.events[i]) +'\n');
	}
	strm.write('\n');
}//writeFileSRT
function eventToSRT(e){
	var sfoo="", ret="", tr={}, bgottext = false;
	//log.debug("eventToVTT: textRuns=["+JSON.stringify(e.textRuns, null, 2)+"]");
	for( var i=0; i<e.textRuns.length; i++ ){
		//if( i==0 ) sfoo = "";
		tr=e.textRuns[i];
		sfoo = tr.text;
		
		if( sfoo.trim().length > 0 )
			bgottext = true;
		//log.debug("eventToVTT: tr.id=["+tr.id+"] bgotext=["+bgottext+"]");
		
		if( tr.linebreak ){
			if( bgottext ){	//linebreak and already found text
				//log.debug("eventToVTT: tr.id=["+tr.id+"] linebreak=true, bgottext=true");
				ret += "\n";
				bgottext = false;
				continue;
			} else {		//linebreak and no text yet
				//log.debug("eventToVTT: tr.id=["+tr.id+"] linebreak=true, bgottext=false");
				continue;
			}
		} else {
		}
		
		//log.debug("eventToSRT: run sfoo=["+sfoo+"] len=["+sfoo.length+"] endval=["+sfoo.charCodeAt(1)+"]");
		//if( sfoo.length==0 ) sfoo = " ";//"\u200B";//use zero-width space instead of a space// =" ";
		if( i==e.textRuns.length-1 ) sfoo = sfoo.replace(/\s+$/, ''); //sfoo = sfoo.trim(); //trim the last space
		
		ret += sfoo;
		//log.debug("eventToSRT: normal ret now=["+ret+"]");
		
	}//textRuns
	
	//log.debug("eventToSRT: finally return ret=["+ret+"] len=["+ret.length+"]");
	return ret;

}//eventToSRT


///////////////////////
// EXPORT TTML
///////////////////////
/*
<?xml version="1.0" encoding="utf-8"?>
<tt xml:lang="en"
    xmlns="http://www.w3.org/ns/ttml"
    xmlns:tts="http://www.w3.org/ns/ttml#styling">
  <head>
    <styling>
      <style id="defaultCaption" tts:fontSize="12px" tts:fontFamily="SansSerif" tts:fontWeight="normal" tts:fontStyle="normal" tts:textDecoration="none" tts:color="white" tts:backgroundColor="black" tts:textAlign="center" />
    </styling>
  </head>
  <body id="thebody" style="defaultCaption">
    <div xml:lang="en">
      <p begin="00:00:07.36" end="00:00:09.63">All:<br/>What, Arthur, what is it?</p>
      <p begin="00:00:17.35" end="00:00:18.30" tts:textAlign="left">Ahem...</p>
      <p begin="00:00:18.30" end="00:00:19.35" tts:textAlign="right">What's the matter?</p>
    </div>
  </body>
</tt>
*/
function writeTTML(fpath, obj){
	//log.debug("writeTTML: enter");
	var wstream = fs.createWriteStream(fpath);
	writeFileTTML(wstream,obj)
	wstream.end();
}//writeTTML
function writeFileTTML(strm, o){
	//log.debug("writeFileTTML: enter");
	//begin header
	strm.write('<?xml version="1.0" encoding="utf-8"?>' + os.EOL);
	strm.write('<tt xml:lang="en" xmlns="http://www.w3.org/ns/ttml" xmlns:tts="http://www.w3.org/ns/ttml#styling">' + os.EOL);
	strm.write('\t<head>' + os.EOL);
	strm.write('\t\t<styling>' + os.EOL);
	strm.write('\t\t\t<style xml:id="defaultCaption" tts:fontSize="12px" tts:fontFamily="SansSerif" tts:fontWeight="normal" tts:fontStyle="normal" tts:textDecoration="none" tts:color="white" tts:backgroundColor="black" tts:textAlign="center" />' + os.EOL);
	strm.write('\t\t</styling>' + os.EOL);
	strm.write('\t</head>' + os.EOL);
	strm.write('\t<body xml:id="thebody" style="defaultCaption">' + os.EOL);
	strm.write('\t\t<div xml:lang="en">' + os.EOL);

	var ev, thisalign;
	for( var i=0; i<o.events.length; i++ ){
		ev = o.events[i];
		
		//set alignment
		//log.debug("writeFileTTML: align=["+ev.align+"]");
		switch (ev.align) {
			case 1:
				thisalign = "left"
				break;
			case 2:
				thisalign = "center"
				break;
			case 3:
				thisalign = "right"
				break;
			default:
				thisalign = "center"
		}//switch align
		
		//write timestamps
		strm.write('\t\t\t<p xml:id="s_'+(i+1)+'" begin="'+o.events[i].intime.toTimecode(2)+'" end="'+o.events[i].outtime.toTimecode(2)+'" tts:textAlign="'+thisalign+'">');
		
		//write textruns
		strm.write(eventToTTML(ev) + '</p>' + os.EOL);
	}
	
	strm.write('\t\t</div>' + os.EOL);
	strm.write('\t</body>' + os.EOL);
	strm.write('</tt>' + os.EOL);
	strm.write(os.EOL);

}//writeFileTTML
function eventToTTML(e){
	//log.debug("eventToTTML: enter");
	var sfoo="", ret="", tr={};
	//log.debug("eventToVTT: textRuns=["+JSON.stringify(e.textRuns, null, 2)+"]");
	for( var i=0; i<e.textRuns.length; i++ ){
		//if( i==0 ) sfoo = "";
		tr=e.textRuns[i];
		sfoo = tr.text; //might be null string if linebreak type
		
		//log.debug("eventToTTML: tr.id=["+tr.id+"] bgotext=["+bgottext+"]");
		
		if( tr.linebreak ){
			ret += ('<br/>');
			continue;
		} else {
		}
		
		//log.debug("eventToTTML: run sfoo=["+sfoo+"] len=["+sfoo.length+"] endval=["+sfoo.charCodeAt(1)+"]");
		//if( sfoo.length==0 ) sfoo = " ";//"\u200B";//use zero-width space instead of a space// =" ";
		if( i==e.textRuns.length-1 ) sfoo = sfoo.replace(/\s+$/, ''); //sfoo = sfoo.trim(); //trim the last space
		//if( tr.bold ) sfoo = sfoo.engulfTTML("b");
		//if( tr.italic ) sfoo = sfoo.engulfTTML("i");
		//if( tr.underline ) sfoo = sfoo.engulfTTML("u");
		//sfoo = sfoo.engulfColorTTML(tr.color,tr.bgcolor);
		
		sfoo = sfoo.engulfTTML(tr);
		
		ret += sfoo;
		//log.debug("eventToTTML: normal ret now=["+ret+"]");
		
	}//textRuns
	
	//log.debug("eventToTTML: finally return ret=["+ret+"] len=["+ret.length+"]");
	return ret;

}//eventToTTML
String.prototype.engulfTTML = function (tr) {
	//wrap the string in <c.fgc.bgc></c>
	var ret = this;
	var stts="";
	
	if( tr.bold ){
		stts += ' tts:fontWeight="bold"';
	}
	if( tr.italic ){
		stts += ' tts:fontStyle="italic"';
	}
	if( tr.underline ){
		stts += ' tts:textDecoration="underline"';
	}
	if( tr.color ){
		stts += (' tts:color="'+rgbToName(tr.color)+'"');
	}
	if( tr.bgcolor ){
		stts += (' tts:backgroundColor="'+rgbToName(tr.bgcolor)+'"');
	}
	
	if( stts.length > 0 )
		ret = '<span'+stts+'>' + this + '</span>';
		
	return( ret ); 
}//String.prototype.engulfTTML


///////////////////////
// EXPORT SAMI
///////////////////////
/*
<SAMI>
<HEAD>
   <STYLE TYPE = "text/css">
   <!--
   // P defines the basic style selector for closed caption paragraph text //
   P {font-family:sans-serif; color:white;}
   // Source, Small, and Big define additional ID selectors for closed caption text //
   #Source {color: orange; font-family: arial; font-size: 12pt;}
   #Small {Name: SmallTxt; font-size: 8pt; color: yellow;}
   #Big {Name: BigTxt; font-size: 12pt; color: magenta;}
   // ENUSCC and FRFRCC define language class selectors for closed caption text //
   .ENUSCC {Name: 'English Captions'; lang: en-US; SAMIType: CC;}
   .FRFRCC {Name: 'French Captions'; lang: fr-FR; SAMIType: CC;}
   -->
   </STYLE>
</HEAD>
<BODY>
   <!<entity type="mdash"/>- The closed caption text displays at 1000 milliseconds. -->
   <SYNC Start = 1000>
      <!-- English closed captions -->
      <P Class = ENUSCC ID = Source>Narrator
      <P Class = ENUSCC>Great reason to visit Seattle, brought to you by two out-of-staters.
      <!-- French closed captions -->
      <P Class = FRFRCC ID = Source>Narrateur
      <P Class = FRFRCC>Deux personnes ne venant la r&eacute;gion vous donnent de bonnes raisons de visiter Seattle.
</BODY>
</SAMI>
*/
function writeSAMI(fpath, obj){
	//log.debug("writeSAMI: enter");
	var wstream = fs.createWriteStream(fpath);
	writeFileSAMI(wstream,obj)
	wstream.end();
}//writeSAMI
function writeFileSAMI(strm, o){
	//log.debug("writeFileSAMI: enter");
	//begin header
	strm.write('<SAMI>' + os.EOL);
	strm.write('\t<HEAD>' + os.EOL);
	strm.write("\t\t<STYLE TYPE='text/css'>" + os.EOL);
	strm.write("\t\t\t<!--" + os.EOL);
	strm.write("\t\t\tP {font-family:sans-serif; color:white; text-align:center;}" + os.EOL);
	strm.write("\t\t\t.ENUSCC {Name: 'English Captions'; lang: en-US; SAMIType: CC;}" + os.EOL);
	strm.write("\t\t\t-->" + os.EOL);
	strm.write('\t\t</StylE>' + os.EOL);
	strm.write('\t</HEAD>' + os.EOL);
	strm.write('\t<BODY>' + os.EOL);

	var ev, thisalign;
	for( var i=0; i<o.events.length; i++ ){
		ev = o.events[i];
		
		//set alignment
		//log.debug("writeFileSAMI: align=["+ev.align+"]");
		switch (ev.align) {
			case 1:
				thisalign = "left"
				break;
			case 2:
				thisalign = "center"
				break;
			case 3:
				thisalign = "right"
				break;
			default:
				thisalign = "center"
		}//switch align
		
		//write timestamps
		strm.write('\t\t<SYNC Start='+Math.floor(o.events[i].intime * 1000)+'>' + os.EOL);
		strm.write('\t\t\t<P Class=ENUSCC>');
		
		//write textruns
		strm.write(eventToSAMI(ev) + '</P>' + os.EOL);
		strm.write('\t\t</SYNC>' + os.EOL);
		
		//write the outtime as a separate caption
		strm.write('\t\t<SYNC Start='+Math.floor(o.events[i].outtime * 1000)+'>' + os.EOL);
		strm.write('\t\t\t<P Class=ENUSCC>&nbsp;<P>' + os.EOL);
		strm.write('\t\t</SYNC>' + os.EOL);

	}//each event
	
	strm.write('\t</BODY>' + os.EOL);
	strm.write('</SAMI>' + os.EOL);
	strm.write(os.EOL);

}//writeFileSAMI
function eventToSAMI(e){
	//log.debug("eventToTTML: enter");
	var sfoo="", ret="", tr={};
	//log.debug("eventToVTT: textRuns=["+JSON.stringify(e.textRuns, null, 2)+"]");
	for( var i=0; i<e.textRuns.length; i++ ){
		//if( i==0 ) sfoo = "";
		tr=e.textRuns[i];
		sfoo = tr.text; //might be null string if linebreak type
		
		//log.debug("eventToTTML: tr.id=["+tr.id+"] bgotext=["+bgottext+"]");
		
		if( tr.linebreak ){
			ret += ('<br/>');
			continue;
		} else {
		}
		
		//log.debug("eventToTTML: run sfoo=["+sfoo+"] len=["+sfoo.length+"] endval=["+sfoo.charCodeAt(1)+"]");
		//if( sfoo.length==0 ) sfoo = " ";//"\u200B";//use zero-width space instead of a space// =" ";
		if( i==e.textRuns.length-1 ) sfoo = sfoo.replace(/\s+$/, ''); //sfoo = sfoo.trim(); //trim the last space
		//if( tr.bold ) sfoo = sfoo.engulfTTML("b");
		//if( tr.italic ) sfoo = sfoo.engulfTTML("i");
		//if( tr.underline ) sfoo = sfoo.engulfTTML("u");
		//sfoo = sfoo.engulfColorTTML(tr.color,tr.bgcolor);
		
		sfoo = sfoo.engulfSAMI(tr);
		
		ret += sfoo;
		//log.debug("eventToTTML: normal ret now=["+ret+"]");
		
	}//textRuns
	
	//log.debug("eventToTTML: finally return ret=["+ret+"] len=["+ret.length+"]");
	return ret;

}//eventToSAMI
String.prototype.engulfSAMI = function (tr) {
	//wrap the string in <c.fgc.bgc></c>
	var ret = this;
	
	return( ret );
	
	var stts="";
	
	if( tr.bold ){
		stts += ' tts:fontWeight="bold"';
	}
	if( tr.italic ){
		stts += ' tts:fontStyle="italic"';
	}
	if( tr.underline ){
		stts += ' tts:textDecoration="underline"';
	}
	if( tr.color ){
		stts += (' tts:color="'+rgbToName(tr.color)+'"');
	}
	if( tr.bgcolor ){
		stts += (' tts:backgroundColor="'+rgbToName(tr.bgcolor)+'"');
	}
	
	if( stts.length > 0 )
		ret = '<span'+stts+'>' + this + '</span>';
		
	return( ret ); 
}//String.prototype.engulfSAMI


///////////////////////
// EXPORT PLAIN TEXT
///////////////////////
/* blank-line-delimited captions

caption here

another caption here


*/
function writePlainText(fpath, obj){
	//log.debug("writePlainText: enter");
	var wstream = fs.createWriteStream(fpath);
	writeFilePlainText(wstream,obj)
	wstream.end();
}//writePlainText
function writeFilePlainText(strm, o){
	//log.debug("writeFilePlainText: enter");

	var ev, thisalign;
	for( var i=0; i<o.events.length; i++ ){
		ev = o.events[i];
		
		//write textruns
		strm.write(os.EOL + eventToPlainText(ev) + os.EOL);
	}
	
}//writeFilePlainText
function eventToPlainText(e){
	//log.debug("eventToPlainText: enter");
	var sfoo="", ret="", tr={};
	//log.debug("eventToPlainText: textRuns=["+JSON.stringify(e.textRuns, null, 2)+"]");
	for( var i=0; i<e.textRuns.length; i++ ){
		//if( i==0 ) sfoo = "";
		tr=e.textRuns[i];
		sfoo = tr.text; //might be null string if linebreak type
		
		//log.debug("eventToVTT: tr.id=["+tr.id+"] bgotext=["+bgottext+"]");
		
		if( tr.linebreak ){
			ret += (os.EOL);
			continue;
		} else {
		}
		
		//log.debug("eventToPlainText: run sfoo=["+sfoo+"] len=["+sfoo.length+"] endval=["+sfoo.charCodeAt(1)+"]");
		//if( sfoo.length==0 ) sfoo = " ";//"\u200B";//use zero-width space instead of a space// =" ";
		if( i==e.textRuns.length-1 ) sfoo = sfoo.replace(/\s+$/, ''); //sfoo = sfoo.trim(); //trim the last space
		//if( tr.bold ) sfoo = sfoo.engulfTTML("b");
		//if( tr.italic ) sfoo = sfoo.engulfTTML("i");
		//if( tr.underline ) sfoo = sfoo.engulfTTML("u");
		//sfoo = sfoo.engulfColorTTML(tr.color,tr.bgcolor);
		
		ret += sfoo;
		//log.debug("eventToPlainText: normal ret now=["+ret+"]");
		
	}//textRuns
	
	//log.debug("eventToPlainText: finally return ret=["+ret+"] len=["+ret.length+"]");
	return ret;

}//eventToPlainText



//project files stored as JSON data
app.post("/openProject/", function(req, res) {
	//log.debug(".post(/openProject/): ENTER");
	//log.debug(".post(/openProject/): stringify data=["+JSON.stringify(req.body)+"]");
	var obj =  JSON.parse(req.body.thedata);
	var projPath = obj.projPath;
	//log.debug(".post(/openProject/): projPath=["+projPath+"]");
	openProject(projPath, res);
})//.post("/openProject/"
function openProject(spath, res){

	/**/
	//log.debug("loadProject: spath=["+spath+"]");
	var projPath = webroot + "/" + spath;
	//log.debug("loadProject: projPath=["+projPath+"]");
	
	fs.readFile( projPath, 'utf8', function (err,data) {
	  if (err) {
	  	log.debug("loadProject: Error reading file=["+projPath+"]");
		return log.debug(err);
	  }
	  //log.debug("loadProject: data=["+data+"]");
	  res.send( data );
	});
	
	
	/*
	var myObject = JSON.parse(fs.createReadStream(projPath));
	strm.read(JSON.stringify(oProject));
	strm.end
	log.debug("loadProject: Stringify=["+JSON.stringify(myObject)+"]");
	*/
	
}//openProject


function rgbToName(c){
	if( c == "#000000" )
		return("black");
	else if( c == "#ffffff" )
		return("white");
	else if( c == "#ff0000" )
		return("red");
	else if( c == "#008000" )
		return("green");
	else if( c == "#0000ff" )
		return("blue");
	else if( c == "#ffff00" )
		return("yellow");
	else if( c == "#00ffff" )
		return("cyan");
	else if( c == "#ff00ff" )
		return("magenta");
	else
		return(c);
}//rgbToName

function itoeAlign(i){
	if( i==ALIGN_LEFT )
		return( "left" );
	else if (i==ALIGN_CENTER )
		return( "center" );
	else if (i==ALIGN_RIGHT )
		return( "right" );
	else
		return( "center" );
}//itoeAlign

function etoiAlign(s){
	if( s=="left" )
		return( ALIGN_LEFT );
	else if (s=="center" )
		return( ALIGN_CENTER );
	else if (s=="right" )
		return( ALIGN_RIGHT );
	else
		return( ALIGN_CENTER );
}//etoiAlign

function itoeVAlign(i){
	if( i==ALIGN_TOP )
		return( "0" );
	else if (i==ALIGN_BOTTOM )
		return( "-1" );
	else
		return( "-1" );
}//itoeVAlign

function etoiVAlign(s){
	if( s=="0" )
		return( ALIGN_TOP );
	else if (s=="-1" )
		return( ALIGN_BOTTOM );
	else
		return( ALIGN_BOTTOM );
}//etoiVAlign

function etoiBold(s){
	if( s=="bold" )
		return( true );
	else
		return( false );
}//etoiBold
function etoiItalic(s){
	if( s=="italic" ||s=="oblique" )
		return( true );
	else
		return( false );
}//etoiItalic
function etoiUnderline(s){
	if( s=="underline" )
		return( true );
	else
		return( false );
}//etoiItalic

function getDateTime() {

    var date = new Date();

    var hour = date.getHours();
    hour = (hour < 10 ? "0" : "") + hour;

    var min  = date.getMinutes();
    min = (min < 10 ? "0" : "") + min;

    var sec  = date.getSeconds();
    sec = (sec < 10 ? "0" : "") + sec;

    var year = date.getFullYear();

    var month = date.getMonth() + 1;
    month = (month < 10 ? "0" : "") + month;

    var day  = date.getDate();
    day = (day < 10 ? "0" : "") + day;

    return year + ":" + month + ":" + day + ":" + hour + ":" + min + ":" + sec;
}//getDateTime



//////////////////////////
// ROUTE IMPORTTYPE
//////////////////////////
app.post("/importtype/:itype", function(req, res) {
    var itype = req.params.itype.toString().toLowerCase();
	//log.debug("app.post(//importtype//): ENTER with itype=["+itype+"]");
    // each key in req.body will match the keys in the data object that you passed in
    var myObject = JSON.parse(req.body.thedata);    
    
	//log.debug("app.post(//importtype//): file name=["+myObject.filename+"]");
	//log.debug("app.post(//importtype//): file dir=["+myObject.dir+"]");
	
    var filepath;
    //webroot starts with a / but shouldn't end with one; mediaDir should begin with one since it comes from the filebrowser
    if( !myObject.dir || !myObject.filename ) return;
    filepath = webroot+myObject.dir + "/" + myObject.filename;
    
    //log.debug("app.post(//importtype//): import type=["+itype+"]");
	switch (itype.toLowerCase()) {
		case "vtt":
			importFileVTT(filepath, res);
			return;
			break;
		case "srt":
			//console.log("SRT import not implemented");
			//return;
			importFileSRT(filepath, res);
			return;
			break;
		case "ttml":
			importFileTTML(filepath, res);
			return;
			break;
		case "qttext":
			importFileQTTEXT(filepath, res);
			return;
			break;
		case "sami":
			console.log("SAMI import not implemented");
			return;
			break;
		case "plain":
			importFilePlain(filepath, res);
			break;
		default:
			console.log("No valid import type specified ["+itype+"].");
			res.status(400).send("Server says Invalid import type specified ["+itype+"]");
			return
	}
    
	return;
});//app.post("/importtype/:itype"


///////////////////////
// IMPORT PLAIN TEXT
///////////////////////
function importFilePlain(filepath, res){

	var ev, myrun, cadetStyle;
	var CadetData = require("./CadetDataNode.js");
	var CadetEvent = require("./CadetEventNode.js");
	var CadetTextRun = require("./CadetTextRunNode.js");
	var CadetStyle = require("./CadetStyleNode.js");
	var data = new CadetData();
		
	var lineReader = require('line-reader');
	linearray = []; //clear the array
	var inacue=false, firsttext=true;
	var thetext="";
	lineReader.eachLine(filepath, function(line, last) {
		//console.log("importFilePlain: line=["+line+"]");
		if( line.trim().length > 0 ){
			//console.log("importFilePlain: found text so inside a cue");
			if( !inacue ){//not already in a cue so create new event
				//data.addEvent("00:00:00.000".timecodeToNumber(),"24:00:00.000".timecodeToNumber());
				data.addEvent("".timecodeToNumber(),"".timecodeToNumber());
				inacue = true;
				firsttext = true;
			}
			
			thetext = line;
			if( firsttext ){
				//console.log("importFilePlain: first text line of cue = ["+thetext+"]");
				
				cadetStyle = new CadetStyle(true, false, false, "white", "black", false); //newline here is false
				textRun = new CadetTextRun( cadetStyle, thetext );
				data.addRun(textRun, false);	//newline indicated as false

				firsttext = false;
				
			} else {	//2nd or more line
				//console.log("importFilePlain: additional text line of cue = ["+thetext+"]");
				//first add blank run with newline
				cadetStyle = new CadetStyle(false, false, false, "white", "black", true);//newline here is true
				textRun = new CadetTextRun( cadetStyle, "" );
				data.addRun(textRun, false);	//newline indicated as true
				//now add the actual text line
				cadetStyle = new CadetStyle(false, false, false, "white", "black", false);//newline here is true
				textRun = new CadetTextRun( cadetStyle, thetext );
				data.addRun(textRun, false);	//newline indicated as true

			}
		} else {
			//console.log("importFilePlain: blank line, end of cue");
			inacue = false;
		}
			
		if (last) {
			var thedata={};
			thedata.data = data;
			res.status(200).send(JSON.stringify(thedata));
			return;
		}
		
	});

	//res.status(400).send("Server says importFilePlain ["+filepath+"] should never get here");
	
}//importFilePlain

/////////////////// 
// IMPORT WEBVTT
///////////////////
// 1
// 00:01:47.430 --> 00:01:51.900   align:center
// <c.white.bgblack>This blade has a dark past.</c>
// 
// 
//var re_time = /(\d+\:\d+\:\d+\.\d+)/;
//var re_time = /-->/;
function importFileVTT(filepath, res){
	var ev, myrun, cadetStyle;
	var CadetData = require("./CadetDataNode.js");
	var CadetEvent = require("./CadetEventNode.js");
	var CadetTextRun = require("./CadetTextRunNode.js");
	var CadetStyle = require("./CadetStyleNode.js");
	var data = new CadetData();
	
	var re_time = /(\d+\:\d+\:\d+\.\d+)\s+-->\s+(\d+\:\d+\:\d+\.\d+)/;
	var re_cuestyle = /(<.*?>)/g;
	var lineReader = require('line-reader');
	linearray = []; //clear the array
	var bfirstline = true, inacue=false, firsttext=true;
	var intime="", outtime="", thetext="";
	lineReader.eachLine(filepath, function(line, last) {
		//console.log("importFileVTT: line=["+line+"]");
		if( bfirstline ){
			//console.log("importFileVTT: first line detected");
			if( line.toLowerCase().search("webvtt") < 0 ){
				res.status(400).send("File ["+filepath+"] doesn't appear to be a WebVTT file.");
				return( false ); //tell line-reader to stop reading and close
			}
		} else {
			if( inacue ){ //check for line of text, either first or not
				//console.log("importFileVTT: inside a cue");
				if( line.trim().length > 0 ){
					//console.log("importFileVTT: found text line");
					thetext = line.replace(re_cuestyle, '');
					if( firsttext ){
						//console.log("importFileVTT: first text line of cue = ["+thetext+"]");
						
						cadetStyle = new CadetStyle(true, false, false, "white", "black", false); //newline here is false
						textRun = new CadetTextRun( cadetStyle, thetext );
						data.addRun(textRun, false);	//newline indicated as false

						firsttext = false;
						
					} else {	//2nd or more line
						//console.log("importFileVTT: additional text line of cue = ["+thetext+"]");
						//first add blank run with newline
						cadetStyle = new CadetStyle(false, false, false, "white", "black", true);//newline here is true
						textRun = new CadetTextRun( cadetStyle, "" );
						data.addRun(textRun, false);	//newline indicated as true
						//now add the actual text line
						cadetStyle = new CadetStyle(false, false, false, "white", "black", false);//newline here is true
						textRun = new CadetTextRun( cadetStyle, thetext );
						data.addRun(textRun, false);	//newline indicated as true

					}
				} else {
					//console.log("importFileVTT: blank line, end of cue");
					inacue = false;
				}
			} else { //check for timecodes, indicating start of cue
				//console.log("importFileVTT: NOT inside a cue");
				if( re_time.test(line) ){	//found a timecode line
					var args = re_time.exec(line);
					intime = args[1];
					outtime = args[2];
					//console.log("importFileVTT: found timecode line in=["+intime+"] out=["+outtime+"]");
					data.addEvent(intime.timecodeToNumber(),outtime.timecodeToNumber());
					//console.log("importFileVTT: event #=["+(data.getEvents().length-1)+"] converted timecode  in=["+intime.timecodeToNumber()+"] out=["+outtime.timecodeToNumber()+"]");
					//(data.getEvents())[data.getEvents().length-1].printTimes();
					//data.printTimes();
					inacue = true;
					firsttext = true; //get ready for first line of cue, probably eating blank lines or other
				} else {
					//console.log("importFileVTT: not in a cue, not a timecode line");
				}
			}
		
		}
		bfirstline = false;
		//return;
	
		//linearray.push(line);
		
		if (last) {
			var thedata={};
			thedata.lines = linearray;
			thedata.data = data;
			//console.log("importFileVTT: eventcount=["+data.getEvents().length+"]");
			//console.log(JSON.stringify(data,null,2));
			//res.status(400).send("server reports importFileVTT found some data");
			res.status(200).send(JSON.stringify(thedata));
			return;
		}
		
	});

	//res.status(400).send("Server says importFileVTT ["+filepath+"] should never get here");
}//importFileVTT

///////////////////////
// IMPORT SRT
///////////////////////
// 1
// 00:01:47.430 --> 00:01:51.900   align:center
// <i>This</i> <b>blade</b> has a <u>dark past</u>.
// 
// 
//var re_time = /(\d+\:\d+\:\d+\.\d+)/;
//var re_time = /-->/;
function importFileSRT(filepath, res){
	var ev, myrun, cadetStyle;
	var CadetData = require("./CadetDataNode.js");
	var CadetEvent = require("./CadetEventNode.js");
	var CadetTextRun = require("./CadetTextRunNode.js");
	var CadetStyle = require("./CadetStyleNode.js");
	var data = new CadetData();
	
	var re_time = /(\d+\:\d+\:\d+[\.\,]\d+)\s+-->\s+(\d+\:\d+\:\d+[\.\,]\d+)/;
	var re_cuestyle = /(<.*?>)/g;
	var lineReader = require('line-reader');
	linearray = []; //clear the array
	var bfirstline = true, inacue=false, firsttext=true;
	var intime="", outtime="", thetext="";
	lineReader.eachLine(filepath, function(line, last) {
		//console.log("importFileSRT: line=["+line+"]");
			if( inacue ){ //check for line of text, either first or not
				//console.log("importFileSRT: inside a cue");
				if( line.trim().length > 0 ){
					//console.log("importFileSRT: found text line");
					thetext = line.replace(re_cuestyle, '');
					if( firsttext ){
						//console.log("importFileSRT: first text line of cue = ["+thetext+"]");
						
						cadetStyle = new CadetStyle(true, false, false, "white", "black", false); //newline here is false
						textRun = new CadetTextRun( cadetStyle, thetext );
						data.addRun(textRun, false);	//newline indicated as false

						firsttext = false;
						
					} else {	//2nd or more line
						//console.log("importFileSRT: additional text line of cue = ["+thetext+"]");
						//first add blank run with newline
						cadetStyle = new CadetStyle(false, false, false, "white", "black", true);//newline here is true
						textRun = new CadetTextRun( cadetStyle, "" );
						data.addRun(textRun, false);	//newline indicated as true
						//now add the actual text line
						cadetStyle = new CadetStyle(false, false, false, "white", "black", false);//newline here is true
						textRun = new CadetTextRun( cadetStyle, thetext );
						data.addRun(textRun, false);	//newline indicated as true

					}
				} else {
					//console.log("importFileSRT: blank line, end of cue");
					inacue = false;
				}
			} else { //check for timecodes, indicating start of cue
				//console.log("importFileSRT: NOT inside a cue");
				if( re_time.test(line) ){	//found a timecode line
					var args = re_time.exec(line);
					intime = args[1].replace(",",".");
					outtime = args[2].replace(",",".");
					//console.log("importFileSRT: found timecode line in=["+intime+"] out=["+outtime+"]");
					data.addEvent(intime.timecodeToNumber(),outtime.timecodeToNumber());
					//console.log("importFileSRT: event #=["+(data.getEvents().length-1)+"] converted timecode  in=["+intime.timecodeToNumber()+"] out=["+outtime.timecodeToNumber()+"]");
					//(data.getEvents())[data.getEvents().length-1].printTimes();
					//data.printTimes();
					inacue = true;
					firsttext = true; //get ready for first line of cue, probably eating blank lines or other
				} else {
					//console.log("importFileSRT: not in a cue, not a timecode line");
				}
			}
		
		
		if (last) {
			var thedata={};
			thedata.data = data;
			//console.log("importFileSRT: eventcount=["+data.getEvents().length+"]");
			//console.log(JSON.stringify(data,null,2));
			//res.status(400).send("server reports importFileVTT found some data");
			res.status(200).send(JSON.stringify(thedata));
			return;
		}
		
	});//forEach

	//res.status(400).send("Server says importFileVTT ["+filepath+"] should never get here");
}//importFileSRT


/////////////////
// IMPORT QTTEXT
/////////////////
// 
// https://developer.apple.com/library/mac/documentation/QuickTime/QT6_3/Chap1/QT6WhatsNew.html
// 
// {QTtext}{font:Geneva}{plain}{size:12}{textColor: 65535, 65535, 65535}{backColor: 0, 0, 0}{justify:center}{timeScale:600}{timeStamps:absolute}{language:0}{textEncoding:0}
// [00:00:00.000]
// Trip to Drumlin Farm
// [00:00:02.000]
// (child noise)
// [00:00:03.000]
// [00:00:04.000]
// Hi kids!
// [00:00:06.000]
// 
function importFileQTTEXT(filepath, res){
	var ev, myrun, cadetStyle;
	var CadetData = require("./CadetDataNode.js");
	var CadetEvent = require("./CadetEventNode.js");
	var CadetTextRun = require("./CadetTextRunNode.js");
	var CadetStyle = require("./CadetStyleNode.js");
	var data = new CadetData();
	
	var re_time = /(\[)(\d+\:\d+\:\d+\.\d+)(\])/; //this pattern:  [00:00:00.000]
	var re_cuestyle = /(\{.*?\})/g;
	var lineReader = require('line-reader');
	var bfirstline = true, inacue=false, firsttext=true;
	var intime="", outtime="", thetext="";
	lineReader.eachLine(filepath, {encoding: 'utf8'}, function(line, last) {
		//console.log("importQTTEXT: line=["+line+"]");
		if( bfirstline ){//check for the sacred qttext marker
			//console.log("importQTTEXT: first line detected firsttwochars=["+line.charCodeAt(0)+":"+line.charCodeAt(1)+":"+line.charCodeAt(3)+"] line=["+line.toLowerCase()+"]");
			if( line.toLowerCase().search("{qttext}") < 0 ){
				res.status(400).send("File ["+filepath+"] doesn't appear to be a QTtext file.");
				return( false ); //tell line-reader to stop reading and close
			}
			bfirstline = false;
		} else {
			//test for timecode line; if yes, update outtime of current event and create new event
			if( re_time.test(line) ){	//found a timecode line
				inacue = true;
				var args = re_time.exec(line);
				intime = args[2];
				outtime = INFINITY;//args[2];  //fix up the outtimes later
				//console.log("importQTTEXT: found timecode line in=["+intime+"] out=["+outtime+"]");
				data.addEvent(intime.timecodeToNumber(),outtime.timecodeToNumber());
				//console.log("importQTTEXT: event #=["+(data.getEvents().length-1)+"] converted timecode  in=["+intime.timecodeToNumber()+"] out=["+outtime.timecodeToNumber()+"]");
				//(data.getEvents())[data.getEvents().length-1].printTimes();
				//data.printTimes();
				firsttext = true; //get ready for first line of cue
			} else if( (line.trim().length > 0) && inacue ){
					//console.log("importQTTEXT: found text line");
					thetext = line.replace(re_cuestyle, '');
					if( firsttext ){
						//console.log("importQTTEXT: first text line of cue = ["+thetext+"]");
						
						cadetStyle = new CadetStyle(true, false, false, "white", "black", false); //newline here is false
						textRun = new CadetTextRun( cadetStyle, thetext );
						data.addRun(textRun, false);	//newline indicated as false

						firsttext = false;
						
					} else {	//2nd or more line
						//console.log("importQTTEXT: additional text line of cue = ["+thetext+"]");
						//first add blank run with newline
						cadetStyle = new CadetStyle(false, false, false, "white", "black", true);//newline here is true
						textRun = new CadetTextRun( cadetStyle, "" );
						data.addRun(textRun, false);	//newline indicated as true
						//now add the actual text line
						cadetStyle = new CadetStyle(false, false, false, "white", "black", false);//newline here is true
						textRun = new CadetTextRun( cadetStyle, thetext );
						data.addRun(textRun, false);	//newline indicated as true

					}//firsttext or not
			} else {
				inacue = false;
			}//timecode or text
		
		}
	
		if (last) {	//fix up all timecodes
		
			//console.log("importQTTEXT: eventcount=["+data.getEvents().length+"]");
			//console.log("importQTTEXT: "+JSON.stringify(data,null,2));
			
			//fix up outTimes; each outtime is next intime (minus blinktime)
			for(var i=0; i<data.getEvents().length; i++){
				var nextintime;
				//console.log("importQTTEXT: i=["+i+"]");
				if( i == data.getEvents().length-1 ){ //last item, set to infinity
						//console.log("importQTTEXT: setting to INFINITY");
						data.events[i].setOuttime(INFINITY.timecodeToNumber());
				} else {	//all other items
					//console.log("importQTTEXT: checking next text length");
					if( data.getText(i+1).trim().length == 0 ){	//next one is an erasure, so use the time as outtime
						nextintime = data.events[i+1].getIntime();
						//console.log("importQTTEXT: setting outtime to nextintime=["+nextintime+"]");
						data.events[i].setOuttime(nextintime);//not otherwise specfied
					} else {	//generic, use -1
						//console.log("importQTTEXT: setting outtime to -1");
						data.events[i].setOuttime(-1);//not otherwise specfied
					}
				}
			}//for all items
			/**/
			// now remove elements with blank captions
			for(var i=data.getEvents().length-1; i>=0; i--){
				if( data.getEvents()[i].getText().trim().length == 0 ){
					data.getEvents().splice(i,1);
				}
			}
			
		
			var thedata={};
			//thedata.lines = linearray;
			thedata.data = data;
			//res.status(400).send("server reports importFileVTT found some data");
			res.status(200).send(JSON.stringify(thedata));
			return;
		}
		
	});

	//res.status(400).send("Server says importQTTEXT ["+filepath+"] should never get here");
}//importFileQTTEXT


/////////////////
// IMPORT TTML
/////////////////
// <?xml version="1.0" encoding="utf-8"?>
// <tt xml:lang="en" xmlns="http://www.w3.org/ns/ttml" xmlns:tts="http://www.w3.org/ns/ttml#styling">
// 	<head>
// 		<styling>
// 			<style id="defaultCaption" tts:fontSize="12px" tts:fontFamily="SansSerif" tts:fontWeight="normal" tts:fontStyle="normal" tts:textDecoration="none" tts:color="white" tts:backgroundColor="black" tts:textAlign="center" />
// 		</styling>
// 	</head>
// 	<body id="thebody" style="defaultCaption">
// 		<div xml:lang="en">
// 			<p id="1" begin="00:01:47.43" end="00:01:51.90" tts:textAlign="center"><span tts:color="red" tts:backgroundColor="black">This blade</span> has a <span tts:color="white" tts:backgroundColor="black">dark past.</span></p>
// 			<p id="2" begin="00:01:51.91" end="00:01:57.99" tts:textAlign="center">It has shed<br>much innocent blood.</p>
// 			<p id="3" begin="00:01:58.00" end="00:02:01.72" tts:textAlign="center">You're a fool for traveling alone<br/>so completely unprepared.</p>
// 		</div>
// 	</body>
// </tt>
// 
var CadetData = require("./CadetDataNode.js");
var CadetEvent = require("./CadetEventNode.js");
var CadetTextRun = require("./CadetTextRunNode.js");
var CadetStyle = require("./CadetStyleNode.js");
var ttmldata;
var initialAlign = ALIGN_CENTER;
var initialStyle={bold:false, italic:false, underline:false, color:"white", bgcolor:"black", linebreak:false};
var align=initialAlign;
function importFileTTML(filepath, res){
	//console.log("importFileTTML: ttmldata=["+JSON.stringify(ttmldata)+"]");
	//var ev, myrun, cadetStyle;
	

	//*****************************************
	//read the file into the data object
	/**/
	fs.readFile(filepath, function(err, filedata) {
	
		if( err ){
			res.status(400).send("Server says importFileTTML ["+filepath+"] failed.");
			return;
		}
		
		//parse the data object into html-ish
		//but first replace malformed self-closing break tags, which Himalaya doesn't handle properly
		//filedata = filedata.toString().replace(/<br\/>/g, "<br \/>");//replace <br/> with <br />
		
		//replace all tabs first
		//filedata = filedata.replace(/\t+/g, "");//replace tabs

		var json = himalaya.parse(filedata);
		//console.log(util.inspect(json[0].children[1], false, null));// [0].body.children[0].div.children[0].p, false, null));

		//var styleOrig = {bold:false, italic:false, underline:false, color:"white", bgcolor:"black", linebreak:false};
		ttmldata = new CadetData();
		//console.log("importFileTTML: calling crawlHimalayaNodes mm:ss=["+getDateTime()+"]") 
		crawlHimalayaNodes(json[0], initialStyle);
		//console.log("importFileTTML: finished crawlHimalayaNodes ms=["+getDateTime()+"]") 
		
		var thedata={};
		//thedata.lines = linearray;
		thedata.data = ttmldata;
		//console.log("importFileVTT: eventcount=["+data.getEvents().length+"]");
		//console.log(JSON.stringify(data,null,2));
		//res.status(400).send("server reports importFileVTT found some data");
		res.status(200).send(JSON.stringify(thedata));
		return;

	});//readfile....
	//*****************************************	
		
}//importFileTTML
var himalayaFoundPara = false;
function crawlHimalayaNodes(n, styleIn){
	//console.log("crawlHimalayaNodes: ttmldata=["+JSON.stringify(ttmldata)+"]");
	//console.log("CRAWL: ENTER");
	//if( n.type && n.type=="Element" ) console.log("CRAWL: ENTER: Element=["+n.tagName+"]");

	var localStyle=extend(true, {},styleIn);//save inbound style to pass to next level; make a deep copy, direct assignment is by reference
			
	if( n.type && (n.type == "Text") && himalayaFoundPara ){
		//console.log("crawlHimalayaNodes--- FOUND Text node and himalayaFoundPara=["+himalayaFoundPara+"]");
		var t = n.content.replace(/\t+/g, "");//replace tabs
		t = t.replace(/\n+/g, "");//replace newlines
		t = t.replace(/\r+/g, "");//replace returns
		if( t.trim().length > 0 ){//ignore newlines and returns
			var thetext = n.content;
			//console.log("crawlHimalayaNodes:--- TEXT: ADD STYLERUN with thetext=["+thetext+"] style=["+JSON.stringify(styleIn)+"]");
			cadetStyle = new CadetStyle(styleIn.bold, styleIn.italic, styleIn.underline, styleIn.color, styleIn.bgcolor, styleIn.linebreak); //newline here is false
			textRun = new CadetTextRun( cadetStyle, thetext );
			ttmldata.addRun(textRun, false);	//newline indicated as false
		}
		himalayaFoundPara = false;//reset
		return;

	} else if( n.tagName && (n.tagName == "tt") ){
		//console.log("crawlHimalayaNodes--- FOUND TT");
		
	} else if( n.tagName && (n.tagName == "head") ){
		//console.log("crawlHimalayaNodes--- FOUND HEAD");
		
	} else if( n.tagName && (n.tagName == "body") ){
		//console.log("crawlHimalayaNodes--- FOUND BODY");
		
	} else if( n.tagName && (n.tagName == "div") ){
		//console.log("crawlHimalayaNodes--- FOUND DIV");
		
	} else if( n.tagName && (n.tagName == "styling") ){
		//console.log("crawlHimalayaNodes--- FOUND STYLING");
		
	} else if( n.tagName && (n.tagName == "style") ){ //maybe grab a default style???
		//console.log("crawlHimalayaNodes--- FOUND STYLE");
		return;
		
	} else if( n.tagName && (n.tagName == "initial") ){ //maybe grab a default style???
		//console.log("crawlHimalayaNodes--- FOUND INITIAL (style)");
		//initialStyle.
		//initialize localStyle
		return;
		
		
	} else if( n.tagName == "p" ){//new caption grab the timecodes and possible textAlign
		//console.log("crawlHimalayaNodes--- FOUND P");
		himalayaFoundPara = true;//signal the next textType node that it's inside a P tag
		var sintime="", souttime="", salign="";
		if( n.attributes.begin ) sintime = n.attributes.begin;
		if( n.attributes.end ) souttime = n.attributes.end;
		if( n.attributes["tts:textAlign"] ) salign = n.attributes["tts:textAlign"];
		if( n.attributes["tts:textAlign"]) {align = etoiAlign(salign);}
		//console.log("crawlHimalayaNodes--- ADD EVENT with intime=["+sintime+"] outtime=["+souttime+"] align=["+salign+"]");
		ttmldata.addEvent(sintime.timecodeToNumber(),souttime.timecodeToNumber(), align);
		align = initialAlign;//reset alignment to initial alignment (which *could* be set from <styling><initial></styling>)

		
	} else if( n.tagName && (n.tagName=="br") ){
		//console.log("crawlHimalayaNodes--- FOUND BR");
		//console.log("--- ADD LINEBREAK StyleRun");
		//cadetStyle = new CadetStyle(false, false, false, "white", "black", true);//newline here is true
		cadetStyle = new CadetStyle(styleIn.bold, styleIn.italic, styleIn.underline, styleIn.color, styleIn.bgcolor, true); //newline here is true
		textRun = new CadetTextRun( cadetStyle, "" );//just a blank run with linebreak set
		//ttmldata.addRun(textRun, true);	//newline indicated as true
		ttmldata.addRun(textRun, false);	//we're handling the linebreak run directly, so no need for data.addRun to handle it
		return;
	} else if( n.tagName && (n.tagName=="br/") ){
		//console.log("crawlHimalayaNodes--- FOUND BR/");
		//console.log("--- ADD LINEBREAK StyleRun");
		//cadetStyle = new CadetStyle(false, false, false, "white", "black", true);//newline here is true
		cadetStyle = new CadetStyle(styleIn.bold, styleIn.italic, styleIn.underline, styleIn.color, styleIn.bgcolor, true); //newline here is true
		textRun = new CadetTextRun( cadetStyle, "" );//just a blank run with linebreak set
		//ttmldata.addRun(textRun, true);	//newline indicated as true
		ttmldata.addRun(textRun, false);	//we're handling the linebreak run directly, so no need for data.addRun to handle it
		//return; //but don't return, Himalaya doesn't handle this quite properly
		
	} else if( n.tagName && (n.tagName == "span") ){//update with new style
		//console.log("crawlHimalayaNodes--- FOUND SPAN");
		if( n.attributes["tts:color"]) {localStyle.color = n.attributes["tts:color"];}
		if( n.attributes["tts:backgroundColor"]) {localStyle.bgcolor = n.attributes["tts:backgroundColor"];}
		if( n.attributes["tts:fontWeight"]) {localStyle.bold = etoiBold(n.attributes["tts:fontWeight"]);}
		if( n.attributes["tts:fontStyle"]) {localStyle.italic = etoiItalic(n.attributes["tts:fontStyle"]);}
		if( n.attributes["tts:textDecoration"]) {localStyle.underline = etoiUnderline(n.attributes["tts:textDecoration"]);}
		//if( n.attributes["tts:textAlign"]) {}//{localStyle.align = etoiAlign(n.attributes["tts:textAlign"]);}
		
	}
	
	//crawl all existing children
	if( n.children ){
		//crawl each child
		n.children.forEach(function(element, index, array){
			crawlHimalayaNodes(element, localStyle);
		});
	}//if there are children
	
}//crawlHimalayaNodes

